/*
 *
 * CTCSS tone generator for Yaesu RH4700
 * code runs on PIC16F628
 *
 * Juha Niinikoski OH2NLT 15.12.2003
 *
 * Sleep and wake up logic added 16.12.2003
 * RH4700 proto & filter tests 19.12.2003
 * Start up indicator added 22.12.2003
*/


#include <pic.h>

/* processor config */

//	__CONFIG(INT & WDTDIS & PWRTEN & BOREN & LVPDIS);	// no WDT
	__CONFIG(XT & MCLRIO & WDTDIS & PWRTEN & BOREN & LVPDIS);	// no WDT

#define RH4700				// generate some RH4700 specific things

/* I/O definitions */

/* Port A */

#define	PORTA_INIT	0		// all filter outputs = 0
#define TRISA_INIT	0x20		// RA5 = input, others = outputs

#define FILTER_PORT	TRISA		// 100n if pin = output, 50n if pin = input (tristate)
#define FILTER_BITS	0x0F		// filter control bits

#define OPERATE		RA4		// operate led
#define ENCODE		RA5		// 1 = generate tones RH4700

/* Port B */

#define	PORTB_INIT	0
#define	TRISB_INIT	0xFD		// RB1 = freq output, others = inputs

#define TONE_BITS	0x3F		// tone select bits
#define TONE_PORT	PORTB

#define TX		RB0		// 1 = TX on (must be RB0 for interrupt)
#define	FO		RB1		// frequency output

/* Other IO */

#define	CMCON_INIT	0x07		// comparators off


/* global variables */

const char id[] = {"Juha Niinikoski OH2NLT 22.12.2003"};

#define LOOP_DELAY 7			// software loop delay us

static char ms10_ctr;
bit ms10_flag;

static char tmr1h;			// set value for TMR1
static char tmr1l;

static int divisor;			// divisor in use

bit test;				// start test


/* Frequency divisor table */
/* for 1 MHz clock (4MHz Xtal) */

const int freq_div[64] = {

	500,	// 0			There are no index 0 selection from RH4700
	7463,	// 1	67		 Yaesu RH4700 basic tones
	6954,	// 2	71,9
	6494,	// 3	77
	6061,	// 4	82,5
	5650,	// 5	88,5
	5274,	// 6	94,8
	5000,	// 7	100
	4831,	// 8	103,5
	4664,	// 9	107,2
	4509,	// 10	110,9
	4355,	// 11	114,8
	4209,	// 12	118,8
	4065,	// 13	123
	3928,	// 14	127,3
	3794,	// 15	131,8
	3663,	// 16	136,3
	3539,	// 17	141,3
	3420,	// 18	146,2
	3303,	// 19	151,4
	3191,	// 20	156,7
	3083,	// 21	162,2
	2978,	// 22	167,9
	2877,	// 23	173,8
	2779,	// 24	179,9
	2685,	// 25	186,2
	2593,	// 26	192,8
	2457,	// 27	203,5
	2373,	// 28	210,7
	2293,	// 29	218,1
	2215,	// 30	225,7
	2140,	// 31	233,6
	2068,	// 32	241,8
	1998,	// 33	250,3

	7463,	// 34	67		 Yaesu high-Q tones
	6954,	// 35	71,9
	6720,	// 36	74,4
	6494,	// 37	77
	6274,	// 38	79,7
	6061,	// 39	82,5
	5855,	// 40	85,4
	5650,	// 41	88,5
	5464,	// 42	91,5
	5133,	// 43	97,4		RH4700 drives only from 1 to 43 tones

	7215,	// 44	69,3		 Other
	3129,	// 45	159,8
	3021,	// 46	165,5
	2919,	// 47	171,3
	2820,	// 48	177,3
	2728,	// 49	183,3
	2633,	// 50	189,9
	2543,	// 51	196.6
	2506,	// 52	199,5
	2421,	// 53	206,5
	2182,	// 54	299,1
	1968,	// 55	254,1

	500,	// 56			Spare 1 KHz tones
	500,	// 57
	500,	// 58
	500,	// 59
	500,	// 60
	500,	// 61
	500,	// 62
	500	// 63
	};

/* Filter capacitor switch table */

const char filter_bits[64] = {
	0,
	0,    0,    0,    0,    0,    0,    0,    0,    0,    0,	//  1 - 10  100n filter cap
	0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08,	// 11 - 20  100n filter cap,  C4 = 50n
	0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F,	// 21 - 30  50n filter cap
	0x0F, 0x0F, 0x0F,						// 31 - 33  50n filter cap
	0,    0,    0,    0,    0,    0,    0,    0,    0,    0,	// 34 - 44  100n filter cap
	0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F,	// 45 - 55  50n filter cap
	0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F			// 56 - 63  50n filter cap
	};


/* Collect bits from I/O and set tone & filter*/
/* Tone bits are RB2-RB7 */

void set_tone(void)
	{
	static char idx;				// divisor table index

	idx = (~TONE_PORT >> 2) & TONE_BITS; 		// RH4700 logic GND = 1, VCC = 0

	divisor = freq_div[idx];
	tmr1l = (0xFFFF - divisor + LOOP_DELAY) & 0xFF;	// 100 Hz test 4 MHz clock
	tmr1h = (0xFFFF - divisor + LOOP_DELAY) / 256;

	FILTER_PORT = TRISA_INIT | filter_bits[idx];	//  set filter

	}


/* Init Interrupt system */

void init_int(void)
	{
	INTEDG = 1;		// set INT rising active
	INTE = 1;
	GIE = 1;		// enable interrupts
	}

/* IRQ handler */

static void interrupt systemint()
	{
	INTF = 0;			// reset IRQ flag
	set_tone();			// set new tone & filter
	}	

/* Alive test */

void alive (void)
	{
	char x;
	TMR1ON = 1;

	for(x=0; x<10; x++)
		{
		while(TMR1IF == 0)
			;
		TMR1IF = 0;	
		OPERATE = test;		// flash led, show we are alive
		test = !test;
		}
	TMR1ON = 0;
	}



/* Main loop */

void main(void)
	{
	CMCON = CMCON_INIT;			// comparators off
	PORTB = PORTB_INIT;
	PORTA = PORTA_INIT;
	TRISB = TRISB_INIT;			// RB5 - RB7 otputs
	TRISA = TRISA_INIT;
#ifdef RH4700
	RBPU = 1;				// B-port pull ups off (TX input shoulg go to 0)
#else
	RBPU = 0;				// B-port pull ups on
#endif

	alive();				// alive indicator

	TMR1L = 0x81;				// 100 Hz before real setup
	tmr1l = 0x81;
	TMR1H = 0x13;
	tmr1h = 0x13;
	TMR1ON = 1;				// start Timer 1

	init_int();				// init interrupt system

	for(;;)					// start for ever loop
		{
		if(ENCODE)			// check if encode tones on
			{
			OPERATE = 0;		// LED on
			while(TX)		// generate tone as long as TX present
				{
				while(TMR1IF == 0)
					;
	
				TMR1H = tmr1h;		// reload timer
				TMR1L = tmr1l;
				TMR1IF = 0;

				FO = !FO;	// generate output
				}
			OPERATE = 1;
			}

		while(!TX)			// idle loop
			{
			asm("sleep");		// sleep while not working
			asm("nop");
			}

		} 
		
	}

